;; prolog.el --- major mode for editing and running Prolog (and Mercury) code

;;; Copyright (C) 1986, 1987, 1997 Free Software Foundation, Inc.
;; Copying and warranty: GNU General Public License, version 2
;;
;; Author: Emil -Astrom-B <emil@sics.se>
;;         Milan Zamazal <pdm@fi.muni.cz>
;;         (see below for more details)
;; Keywords: prolog major mode sicstus swi mercury
;; Version: 0.1.12

;;
;; Original author: Masanobu UMEDA <umerin@mse.kyutech.ac.jp>
;; Parts of this file was taken from a modified version of the original
;; by Johan Andersson, Peter Olin, Mats Carlsson, Johan Bevemyr, Stefan
;; Andersson, and Per Danielsson (all SICS people), and Henrik B-Akman-B
;; at Uppsala University, Sweden.
;;
;; Some ideas and also a few lines of code have been borrowed (not stolen ;-)
;; from Oz.el, the Emacs major mode for the Oz programming language,
;; Copyright (C) 1993 DFKI GmbH, Germany, with permission.
;; Authors: Ralf Scheidhauer and Michael Mehl ([scheidhr|mehl]@dfki.uni-sb.de)

;;; Commentary:
;;
;; This package provides a major mode for editing Prolog.  It knows
;; about Prolog syntax and comments, and can send regions to an inferior
;; Prolog interpreter process. Font-lock and auto indentation is
;; supported.
;;
;; The code requires the comint, easymenu, info, imenu, and font-lock
;; libraries. These are normally distributed with GNU Emacs.

;;; Installation:
;;
;; Insert the following lines in your ~/.emacs (or _emacs in case
;; of Windows/DOS) to make Emacs use this mode when editing Prolog files:
;;
;; (setq load-path (cons "/usr/local/lib/sicstus3" load-path))
;; (autoload 'run-prolog "prolog" "Start a Prolog sub-process." t)
;; (autoload 'prolog-mode "prolog" "Major mode for editing Prolog programs." t)
;; (autoload 'mercury-mode "prolog" "Major mode for editing Mercury programs." t)
;; (setq prolog-system 'sicstus)
;; (setq auto-mode-alist (append '(("\\.pl$" . prolog-mode)
;;                                 ("\\.m$" . mercury-mode))
;;                                auto-mode-alist))
;;
;; where the path in the first line is the file system path to this file.
;; MSDOS paths can be written like "d:/programs/emacs-19.34/site-lisp". 
;;
;; The last line above makes sure that files ending with .pl are
;; assumed to be Prolog files and not Perl, which is the default Emacs
;; setting. If this is not wanted, remove that line. It is then necessary
;; for the user to manually switch to prolog mode by typing
;; M-x prolog-mode after opening a prolog file.
;;
;; If the shell command to start the prolog process ('sicstus' or 'pl' for SWI
;; prolog) is not available in the default path, then it is necessary to
;; set the value of the environment variable EPROLOG to a shell command
;; to invoke the prolog process.

;;; Known problems:
;;
;; o Quoted atoms are not strings in the syntax table. This means that
;;   minor modes such as the show-paren-mode will sometimes behave
;;   strangely since they match on characters inside atoms.

;;; Future work:
;;
;; o Caching of neutral points to speed up parsing (and thus indentation)
;; o If syntax error in SICStus, then we could parse the code to find
;;   the exact location.


;;; Code:

;(eval-when-compile
;  (require 'comint)
;  (require 'easymenu)
;  (require 'info)
;  )


(require 'comint)
(require 'easymenu)

(defgroup prolog nil
  "Major mode for editing and running Prolog under Emacs."
  :group 'languages)

(defgroup prolog-faces nil
  "Prolog specific faces."
  :group 'font-lock)

;;-------------------------------------------------------------------
;; User configurable variables
;;-------------------------------------------------------------------

;; General configuration

(defcustom prolog-system nil
  "*Prolog interpreter/compiler used.
The value of this variable is nil or a symbol.
If it is a symbol, it determines default values of other configuration
variables with respect to properties of the specified Prolog
interpreter/compiler.

Currently recognized symbol values are:
mercury - Mercury
sicstus - SICStus Prolog
swi     - SWI Prolog"
  :group 'prolog
  :type 'symbol
  :options '(nil sicstus swi)) ; not mercury, bec. it's set in `mercury-mode'
(make-variable-buffer-local 'prolog-system)

;; Indentation

(defcustom prolog-indent-width tab-width
  "*The indentation width used by the editing buffer."
  :group 'prolog
  :type 'integer)  

(defcustom prolog-align-comments-flag t
  "*Non-nil means automatically align comments when indenting."
  :group 'prolog
  :type 'boolean)

(defcustom prolog-indent-mline-comments-flag t
  "*If non-nil, indent contents of /* */ comments.
Otherwise leave such lines as they are."
  :group 'prolog
  :type 'boolean)

(defcustom prolog-object-end-to-0-flag t
  "*If non-nil, indent closing '}' in object definitions to level 0.
Otherwise indent to `prolog-indent-width'."
  :group 'prolog
  :type 'boolean)

(defcustom prolog-left-indent-regexp "\\(;\\|\\*?->\\)"
  "*Regexp for character sequences after which next line is indented.
Next line after such a regexp is indented to the opening paranthesis level."
  :group 'prolog
  :type 'regexp)

(defcustom prolog-paren-indent 4
  "*The indentation increase for parenthesis expressions.
Only used in ( If -> Then ; Else) and ( Disj1 ; Disj2 ) style expressions."
  :group 'prolog
  :type 'integer)

(defcustom prolog-parse-mode 'beg-of-line
  "*The parse mode used (decides from which point parsing is done).
Legal values:
'beg-of-line   - starts parsing at the beginning of a line, unless the
                 previous line ends with a backslash. Fast, but has
                 problems detecting multiline /* */ comments.
'beg-of-clause - starts parsing at the beginning of the current clause.
                 Slow, but copes better with /* */ comments."
  :group 'prolog
  :type 'symbol
  :options '(beg-of-line beg-of-clause))

;; Font locking

(defcustom prolog-keywords
  '((mercury
     ("all" "else" "end_module" "equality" "external" "fail" "func" "if"
      "implementation" "import_module" "inst" "interface" "mode" "module" "not"
      "pragma" "pred" "some" "then" "true" "type" "use_module" "where"))
    (sicstus
     ("block" "dynamic" "mode" "module" "multifile" "meta_predicate"
      "parallel" "public" "sequential"))
    (swi
     ("discontiguous" "dynamic" "ensure_loaded" "export" "export_list" "import"
      "meta_predicate" "module" "module_transparent" "multifile" "require"
      "use_module" "volatile"))
    (t
     ("dynamic" "module")))
  "*Alist of Prolog keywords which is used for font locking of directives."
  :group 'prolog
  :type 'sexp)

(defcustom prolog-types
  '((mercury
     ("char" "float" "int" "io__state" "string" "univ"))
    (t nil))
  "*Alist of Prolog types used by font locking."
  :group 'prolog
  :type 'sexp)

(defcustom prolog-mode-specificators
  '((mercury
     ("bound" "di" "free" "ground" "in" "mdi" "mui" "muo" "out" "ui" "uo"))
    (t nil))
  "*Alist of Prolog mode specificators used by font locking."
  :group 'prolog
  :type 'sexp)

(defcustom prolog-determinism-specificators
  '((mercury
     ("cc_multi" "cc_nondet" "det" "erroneous" "failure" "multi" "nondet"
      "semidet"))
    (t nil))
  "*Alist of Prolog determinism specificators used by font locking."
  :group 'prolog
  :type 'sexp)

(defcustom prolog-directives
  '((mercury
     ("^#[0-9]+"))
    (t nil))
  "*Alist of Prolog source code directives used by font locking."
  :group 'prolog
  :type 'sexp)


;; Keyboard

(defcustom prolog-electric-newline-flag t
  "*If non-nil, automatically indent the next line when the user types RET."
  :group 'prolog
  :type 'boolean)

(defcustom prolog-hungry-delete-key-flag nil
  "*If non-nil, delete key consumes all preceding spaces."
  :group 'prolog
  :type 'boolean)

(defcustom prolog-electric-dot-flag nil
  "*If non-nil, dot appends newline or inserts head of a new clause.
If dot is pressed at the end of nonempty line, it inserts dot and newline.
If dot is pressed at the empty line, it inserts head of a new clause for the
last predicate. 
It does not apply in strings and comments."
  :group 'prolog
  :type 'boolean)

(defcustom prolog-old-sicstus-keys-flag nil
  "*If non-nil, old SICStus Prolog mode keybindings are used."
  :group 'prolog
  :type 'boolean)

;; Inferior mode

(defcustom prolog-program-name
  '(((getenv "EPROLOG") (eval (getenv "EPROLOG")))
    (sicstus "sicstus")
    (swi "pl")
    (mercury nil)
    (t "prolog"))
  "*Alist of program names for invoking an inferior Prolog with `run-prolog'."
  :group 'prolog
  :type 'sexp)

(defcustom prolog-program-switches
  '((sicstus ("-i"))
    (t nil))
  "*Alist of switches given to inferior Prolog run with `run-prolog'."
  :group 'prolog
  :type 'sexp)

(defcustom prolog-consult-string
  '((sicstus "prolog:zap_file(%m,%b,consult).")
    (swi "[%f].")
    (mercury nil)
    (t "reconsult(%f)."))
  "*Alist of strings defining predicate for reconsulting.
Some parts of the string are replaced:
`%f' by the name of the consulted file (can be a temporary file)
`%b' by the file name of the buffer to consult
`%m' by the module name and name of the consulted file separated by colon"
  :group 'prolog
  :type 'sexp)

(defcustom prolog-compile-string
  '((sicstus "prolog:zap_file(%m,%b,compile).")
    (swi "[%f].")
    (mercury "mmake ")
    (t "compile(%f)."))
  "*Alist of strings and lists defining predicate for recompilation.

Some parts of the string are replaced:
`%f' by the name of the compiled file (can be a temporary file)
`%b' by the file name of the buffer to compile
`%m' by the module name and name of the compiled file separated by colon

If `prolog-program-name' is non-nil, it is a string sent to a Prolog process.
If `prolog-program-name' is nil, it is an argument to the `compile' function."
  :group 'prolog
  :type 'sexp)

(defcustom prolog-eof-string "end_of_file.\n"
  "*Alist of strings that represent end of file for prolog.
nil means send actual operating system end of file."
  :group 'prolog
  :type 'sexp)

(defcustom prolog-prompt-regexp
  '((sicstus "^| [ ?][- ] *")
    (swi "^[1-9][0-9]* \\?- \\|^| +")
    (t "^ *\\?-"))
  "*Alist of prompts of the prolog system command line."
  :group 'prolog
  :type 'sexp)

(defcustom prolog-continued-prompt-regexp
  '((sicstus "^\\(| +\\|     +\\)")
    (t "^|: +"))
  "*Alist of regexps matching the prompt when consulting `user'."
  :group 'prolog
  :type 'sexp)

(defcustom prolog-debug-on-string "debug.\n"
  "*Predicate for enabling debug mode."
  :group 'prolog
  :type 'string)

(defcustom prolog-debug-off-string "nodebug.\n"
  "*Predicate for disabling debug mode."
  :group 'prolog
  :type 'string)

(defcustom prolog-trace-on-string "trace.\n"
  "*Predicate for enabling tracing."
  :group 'prolog
  :type 'string)

(defcustom prolog-trace-off-string "notrace.\n"
  "*Predicate for disabling tracing."
  :group 'prolog
  :type 'string)

(defcustom prolog-use-standard-consult-compile-method-flag t
  "*If non-nil use the standard compilation method.
Otherwise the new compilation method will be used. This
utilises a special compilation buffer with the associated
features such as parsing of error messages and automatically
jumping to the source code responsible for the error.

Warning: the new method is so far only experimental and
does contain bugs. The recommended setting for the novice user
is non-nil for this variable."
  :group 'prolog
  :type 'boolean)


;; Miscellaneous

(defcustom prolog-use-prolog-tokenizer-flag t
  "*If non-nil, then use the internal prolog tokenizer for indentation etc.
Otherwise use `parse-partial-sexp' which is faster but sometimes incorrect."
  :group 'prolog
  :type 'boolean)

(defcustom prolog-imenu-flag nil
  "*If non-nil, then add a clause index menu for all prolog files."
  :group 'prolog
  :type 'boolean)

(defcustom prolog-info-predicate-index
  "(sicstus)Predicate Index"
  "*The info node for the SICStus predicate index."
  :group 'prolog
  :type 'string)



;;-------------------------------------------------------------------
;; Internal variables
;;-------------------------------------------------------------------

(defvar prolog-emacs 
  (if (string-match "XEmacs\\|Lucid" emacs-version)
      'xemacs
    'gnuemacs)
  "The variant of Emacs we're running.
Valid values are 'gnuemacs and 'xemacs.")

(defvar prolog-known-systems '(mercury sicstus swi))

(defvar prolog-temp-filename "")   ; Later set by `prolog-temporary-file'

(defvar prolog-mode-syntax-table nil)
(defvar prolog-mode-abbrev-table nil)
(defvar prolog-mode-map nil)
  
(defconst prolog-left-paren "[[({]" 
  "The characters that are used as left parentheses for the indentation code.")
(defconst prolog-right-paren "[])}]"
  "The characters that are used as right parentheses for the indentation code.")
(defconst prolog-atom-char-regexp "[a-zA-Z0-9_$]"
  "Regexp specifying characters which constitute atoms without quoting.")
(defconst prolog-atom-regexp (format "[a-z$]+%s*" prolog-atom-char-regexp)
  "Regexp specifying non-quoted atoms.")
(defconst prolog-quoted-atom-regexp
  "\\(^\\|[^0-9]\\)\\('\\([^\n']\\|\\\\'\\)*'\\)"
;  "[^0-9]\\('\\([^\n']\\|\\\\'\\)*'\\)"
  "Regexp matching a quoted atom.")
(defconst prolog-string-regexp
  "\\(\"\\([^\n\"]\\|\\\\\"\\)*\"\\)"
  "Regexp matching a string.")
(defconst prolog-head-delimiter "\\(:-\\|\\+:\\|-:\\|\\+\\?\\|-\\?\\|-->\\)"
  "A regexp for matching on the end delimiter of a head (e.g. \":-\").")

(defvar prolog-compilation-buffer "*prolog-compilation*"
  "Name of the output buffer for Prolog compilation/consulting.")


;;-------------------------------------------------------------------
;; Prolog mode
;;-------------------------------------------------------------------

(if prolog-mode-syntax-table
    ()
  (let ((table (make-syntax-table)))
    (modify-syntax-entry ?_ "_" table)
    (modify-syntax-entry ?* ". 23b" table)
    (modify-syntax-entry ?/ ". 124" table)
    (modify-syntax-entry ?+ "." table)
    (modify-syntax-entry ?- "." table)
    (modify-syntax-entry ?= "." table)
    (modify-syntax-entry ?% "<" table)
    (modify-syntax-entry ?\n ">" table)
    (modify-syntax-entry ?< "." table)
    (modify-syntax-entry ?> "." table)
    (modify-syntax-entry ?| "." table)
    (modify-syntax-entry ?\' "." table)
    (setq prolog-mode-syntax-table table)))

(define-abbrev-table 'prolog-mode-abbrev-table ())

(defun prolog-find-value-by-system (alist)
  "Get value from ALIST according to `prolog-system'."
  (if (listp alist)
      (let (result
	    id)
	(while alist
	  (setq id (car (car alist)))
	  (if (or (eq id prolog-system)
		  (eq id t)
		  (and (listp id)
		       (eval id)))
	      (progn
		(setq result (car (cdr (car alist))))
		(if (and (listp result)
			 (eq (car result) 'eval))
		    (setq result (eval (car (cdr result)))))
		(setq alist nil))
	    (setq alist (cdr alist))))
	result)
    alist))

(defun prolog-mode-variables ()
  "Set some common variables to Prolog code specific values."
  (setq local-abbrev-table prolog-mode-abbrev-table)
  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "[ \t]*$\\|" page-delimiter)) ;'%%..'
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'paragraph-ignore-fill-prefix)
  (setq paragraph-ignore-fill-prefix t)
  (make-local-variable 'adaptive-fill-mode)
  (setq adaptive-fill-mode nil)
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'prolog-indent-line)
  (make-local-variable 'comment-start)
  (setq comment-start "%")
  (make-local-variable 'comment-end)
  (setq comment-end "")
  (make-local-variable 'comment-start-skip)
  ;; This complex regexp makes sure that comments cannot start
  ;; inside quoted atoms or strings
  (setq comment-start-skip 
	(format "^\\(\\(%s\\|%s\\|[^\n\'\"%%]\\)*\\)\\(/\\*+ *\\|%%+ *\\)" 
		prolog-quoted-atom-regexp prolog-string-regexp))
  (make-local-variable 'comment-column)
  (make-local-variable 'comment-indent-function)
  (setq comment-indent-function 'prolog-comment-indent)
  (make-local-variable 'comment-indent-function)
  (setq comment-indent-function 'prolog-comment-indent)
  ;; Initialize Prolog system specific variables
  (let ((vars '(prolog-keywords prolog-types prolog-mode-specificators
		prolog-determinism-specificators prolog-directives
		prolog-program-name prolog-program-switches
		prolog-consult-string prolog-compile-string prolog-eof-string
		prolog-prompt-regexp prolog-continued-prompt-regexp
		prolog-help-function)))
    (while vars
      (set (intern (concat (symbol-name (car vars)) "-i"))
	   (prolog-find-value-by-system (eval (car vars))))
      (setq vars (cdr vars))))
  (if (null prolog-program-name-i)
      (setq compile-command prolog-compile-string-i))
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults
	'(prolog-font-lock-keywords nil nil ((?_ . "w"))))
)

(defun prolog-mode-keybindings-common (map)
  "Define keybindings common to both Prolog modes."
  (define-key map "\C-c?" 'prolog-help-on-predicate)
  (define-key map "\C-c/" 'prolog-help-apropos)
  (define-key map "\C-c\C-d" 'prolog-debug-on)
  (define-key map "\C-c\C-t" 'prolog-trace-on)
  (define-key map "\C-c\r" 'run-prolog))

(defun prolog-mode-keybindings-edit (map)
  "Define keybindings for prolog-mode."
  (define-key map "\M-{" 'prolog-goto-clause-start)
  (define-key map "\M-}" 'prolog-goto-clause-end)
  (define-key map "\M-\C-c" 'prolog-mark-clause)
  (define-key map "\C-\M-a" 'prolog-goto-pred-start)
  (define-key map "\C-\M-e" 'prolog-goto-pred-end)
  (define-key map "\M-\C-h" 'prolog-mark-predicate)
  (define-key map "\M-\C-n" 'prolog-forward-list)
  (define-key map "\M-\C-p" 'prolog-backward-list)
  (define-key map "\C-c\C-n" 'prolog-insert-predicate-template)
  (define-key map "\C-c\C-s" 'prolog-insert-predspec)
  (define-key map "\M-\r" 'prolog-insert-next-clause)

  (define-key map "\177" 'prolog-electric-delete)
  (define-key map "." 'prolog-electric-dot)
  (if prolog-electric-newline-flag 
      (define-key map "\r" 'newline-and-indent))

  (if prolog-old-sicstus-keys-flag
      (progn
	(define-key map "\C-c\C-c" 'prolog-consult-predicate)
	(define-key map "\C-cc" 'prolog-consult-region)
	(define-key map "\C-cC" 'prolog-consult-buffer)
	(define-key map "\C-c\C-k" 'prolog-compile-predicate)
	(define-key map "\C-ck" 'prolog-compile-region)
	(define-key map "\C-cK" 'prolog-compile-buffer))
    (define-key map "\C-c\C-p" 'prolog-consult-predicate)
    (define-key map "\C-c\C-r" 'prolog-consult-region)
    (define-key map "\C-c\C-b" 'prolog-consult-buffer)
    (define-key map "\C-c\C-f" 'prolog-consult-file)
    (define-key map "\C-c\C-cp" 'prolog-compile-predicate)
    (define-key map "\C-c\C-cr" 'prolog-compile-region)
    (define-key map "\C-c\C-cb" 'prolog-compile-buffer)
    (define-key map "\C-c\C-cf" 'prolog-compile-file)))

(defun prolog-mode-keybindings-inferior (map)
  "Define keybindings for prolog-inferior-mode."
  ;; No inferior mode specific keybindings now.
  )

(if prolog-mode-map
    ()
  (setq prolog-mode-map (make-sparse-keymap))
  (prolog-mode-keybindings-common prolog-mode-map)
  (prolog-mode-keybindings-edit prolog-mode-map)
  ;; System dependent keymaps for system dependent menus
  (let ((systems prolog-known-systems))
    (while systems
; Why do we have the mode-map as a system-specific variable?
      (set (intern (concat "prolog-mode-map-" (symbol-name (car systems))))
	   (cons 'keymap prolog-mode-map))
;      (set (intern (concat "prolog-mode-map-" (symbol-name (car systems))))
;	   prolog-mode-map)
      (setq systems (cdr systems)))))
  

(defvar prolog-mode-hook nil)

;;;###autoload
(defun prolog-mode (&optional system)
  "Major mode for editing Prolog code.

Blank lines and `%%...' separate paragraphs.  `%'s starts a comment
line and comments can also be enclosed in /* ... */.

Commands:
\\{prolog-mode-map}
Entry to this mode calls the value of `prolog-mode-hook'
if that value is non-nil."
  (interactive)
  (kill-all-local-variables)
  (if system (setq prolog-system system))
  (use-local-map
   (if prolog-system
       (eval (intern (concat "prolog-mode-map-" (symbol-name prolog-system))))
     prolog-mode-map)
;   prolog-mode-map
   )
  (setq major-mode 'prolog-mode)
  (setq mode-name (concat "Prolog"
			  (cond
			   ((eq prolog-system 'mercury) "[Mercury]")
			   ((eq prolog-system 'sicstus) "[SICStus]")
			   ((eq prolog-system 'swi) "[SWI]")
			   (t ""))))
  (set-syntax-table prolog-mode-syntax-table)
  (prolog-mode-variables)

  ;; Add predicate index menu
  (make-variable-buffer-local 'imenu-create-index-function)
  (setq imenu-create-index-function 'imenu-default-create-index-function)
;Emil (does not detect facts, problems in Windows? Try TACIT main.pl!)
;  (setq imenu-create-index-function 'prolog-create-predicate-index)
;Milan (this has problems with object methods...)
  (setq imenu-prev-index-position-function 'prolog-goto-pred-start)
  (setq imenu-extract-index-name-function 'prolog-get-predspec)

  (if prolog-imenu-flag
      (imenu-add-to-menubar "Predicates"))

  (run-hooks 'prolog-mode-hook))

;;;###autoload
(defun mercury-mode ()
  "Major mode for editing Mercury programs.
Actually this is just customized prolog-mode, so see `prolog-mode'
documentation for more information."
  (interactive)
  (prolog-mode 'mercury))


;;-------------------------------------------------------------------
;; Inferior prolog mode
;;-------------------------------------------------------------------

(defvar prolog-inferior-mode-map nil)
(defvar prolog-inferior-mode-hook nil)

(defun prolog-inferior-mode ()
  "Major mode for interacting with an inferior Prolog process.

The following commands are available:
\\{prolog-inferior-mode-map}

Entry to this mode calls the value of prolog-mode-hook with no arguments,
if that value is non-nil.  Likewise with the value of comint-mode-hook.
prolog-mode-hook is called after comint-mode-hook.

You can send text to the inferior Prolog from other buffers
using the commands send-region, send-string and \\[prolog-consult-region].

Commands:
Tab indents for Prolog; with argument, shifts rest
 of expression rigidly with the current line.
Paragraphs are separated only by blank lines and '%%'. '%'s start comments.

Return at end of buffer sends line as input.
Return not at end copies rest of line to end and sends it.
\\[comint-delchar-or-maybe-eof] sends end-of-file as input.
\\[comint-kill-input] and \\[backward-kill-word] are kill commands, imitating normal Unix input editing.
\\[comint-interrupt-subjob] interrupts the shell or its current subjob if any.
\\[comint-stop-subjob] stops, likewise. \\[comint-quit-subjob] sends quit signal, likewise."
  (interactive)
  (cond ((not (eq major-mode 'prolog-inferior-mode))
	 (kill-all-local-variables)
	 (comint-mode)
	 (setq comint-input-filter 'prolog-input-filter)
	 (setq major-mode 'prolog-inferior-mode)
	 (setq mode-name "Inferior Prolog")
	 (setq mode-line-process '(": %s"))
	 (prolog-mode-variables)
	 (if prolog-inferior-mode-map
	     ()
	   (setq prolog-inferior-mode-map (copy-keymap comint-mode-map))
	   (prolog-mode-keybindings-common prolog-inferior-mode-map)
	   (prolog-mode-keybindings-inferior prolog-inferior-mode-map))
	 (use-local-map prolog-inferior-mode-map)
	 (setq comint-prompt-regexp prolog-prompt-regexp-i)
	 (make-variable-buffer-local 'shell-dirstack-query)
	 (setq shell-dirstack-query "pwd.")
	 (run-hooks 'prolog-inferior-mode-hook))))

(defun prolog-input-filter (str)
  (cond ((string-match "\\`\\s *\\'" str) nil) ;whitespace
	((not (eq major-mode 'prolog-inferior-mode)) t)
	((= (length str) 1) nil)	;one character
	((string-match "\\`[rf] *[0-9]*\\'" str) nil) ;r(edo) or f(ail)
	(t t)))

;;;###autoload
(defun run-prolog (prefix)
  "Run an inferior Prolog process, input and output via buffer *prolog*.
With prefix argument, restart the Prolog process if running before."
  (interactive "P")
  (if (and prefix (get-process "prolog"))
      (progn
	(process-send-string "prolog" "halt.\n")
	(while (get-process "prolog") (sit-for 0.1))))
  (let ((buff (buffer-name)))
    (if (not (string= buff "*prolog*"))
;	(switch-to-buffer-other-window "*prolog*"))
	(prolog-goto-prolog-process-buffer))
    (prolog-mode-variables)
    (prolog-ensure-process)))

(defun prolog-ensure-process (&optional wait)
  "If Prolog process is not running, run it.
If the optional argument WAIT is non-nil, wait for Prolog prompt specified by
the variable `prolog-prompt-regexp'."
  (if (null prolog-program-name-i)
      (error "This Prolog system has defined no interpreter."))
  (if (comint-check-proc "*prolog*")
      ()
    (apply 'make-comint "prolog" prolog-program-name-i nil
	   prolog-program-switches-i)
    (save-excursion
      (set-buffer "*prolog*")
      (prolog-inferior-mode)
      (if wait
	  (progn
	    (goto-char (point-max))
	    (while
		(save-excursion
		  (not
		   (re-search-backward
		    (concat "\\(" prolog-prompt-regexp-i "\\)" "\\=")
		    nil t)))
	      (sit-for 0.1)))))))

(defun prolog-process-insert-string (process string)
  "Insert STRING into inferior Prolog buffer running PROCESS."
  ;; Copied from elisp manual, greek to me
  (let ((buf (current-buffer)))
    (unwind-protect
	(let (moving)
	  (set-buffer (process-buffer process))
	  (setq moving (= (point) (process-mark process)))
	  (save-excursion
	    ;; Insert the text, moving the process-marker.
	    (goto-char (process-mark process))
	    (insert string)
	    (set-marker (process-mark process) (point)))
	  (if moving (goto-char (process-mark process))))
      (set-buffer buf))))


;;------------------------------------------------------------
;; Old consulting and compiling functions
;;------------------------------------------------------------

(defun prolog-old-process-region ( compilep start end)
  "Process the region limited by START and END positions.
If COMPILEP is non-nil then use compilation, otherwise consulting."
   (prolog-ensure-process)
   (let ((tmpfile prolog-temp-filename)
	 (process (get-process "prolog")))
     (write-region start end tmpfile)
     (process-send-string
      "prolog" (prolog-build-prolog-command
		compilep tmpfile (prolog-bsts buffer-file-name)))
;     (switch-to-buffer-other-window "*prolog*")))
     (prolog-goto-prolog-process-buffer)))

(defun prolog-old-process-predicate ( compilep )
  "Process the predicate around point.
If COMPILEP is non-nil then use compilation, otherwise consulting."
  (prolog-old-process-region
   compilep (prolog-pred-start) (prolog-pred-end)))

(defun prolog-old-process-buffer ( compilep )
  "Process the entire buffer.
If COMPILEP is non-nil then use compilation, otherwise consulting."
  (prolog-old-process-region compilep (point-min) (point-max)))

(defun prolog-old-process-file ( compilep )
  "Process the file of the current buffer.
If COMPILEP is non-nil then use compilation, otherwise consulting."
  (save-some-buffers)
  (prolog-ensure-process)
  (let ((filename (prolog-bsts buffer-file-name)))
    (process-send-string
     "prolog" (prolog-build-prolog-command 
	       compilep filename filename))
;    (switch-to-buffer-other-window "*prolog*")))
    (prolog-goto-prolog-process-buffer)))

;;------------------------------------------------------------
;; Consulting and compiling
;;------------------------------------------------------------

;;; Interactive interface functions, used by both the standard
;;; and the experimental consultation and compilation functions
(defun prolog-consult-file ()
  "Consult file of current buffer."
  (interactive)
  (if prolog-use-standard-consult-compile-method-flag
      (prolog-old-process-file nil)
    (prolog-consult-compile-file nil)))

(defun prolog-consult-buffer ()
  "Consult buffer."
  (interactive)
  (if prolog-use-standard-consult-compile-method-flag
      (prolog-old-process-buffer nil)
    (prolog-consult-compile-buffer nil)))

(defun prolog-consult-region (beg end)
  "Consult region."
  (interactive "r")
  (if prolog-use-standard-consult-compile-method-flag
      (prolog-old-process-region nil beg end)
    (prolog-consult-compile-region nil beg end)))

(defun prolog-consult-predicate ()
  "Consult the predicate around current point."
  (interactive)
  (if prolog-use-standard-consult-compile-method-flag
      (prolog-old-process-predicate nil)
    (prolog-consult-compile-predicate nil)))

(defun prolog-compile-file ()
  "Compile file of current buffer."
  (interactive)
  (if prolog-use-standard-consult-compile-method-flag
      (prolog-old-process-file t)
    (prolog-consult-compile-file t)))

(defun prolog-compile-buffer ()
  "Compile buffer."
  (interactive)
  (if prolog-use-standard-consult-compile-method-flag
      (prolog-old-process-buffer t)
    (prolog-consult-compile-buffer t)))

(defun prolog-compile-region (beg end)
  "Compile region."
  (interactive "r")
  (if prolog-use-standard-consult-compile-method-flag
      (prolog-old-process-region t beg end)
    (prolog-consult-compile-region t beg end)))

(defun prolog-compile-predicate ()
  "Compile the predicate around current point."
  (interactive)
  (if prolog-use-standard-consult-compile-method-flag
      (prolog-old-process-predicate t)
    (prolog-consult-compile-predicate t)))

(defun prolog-buffer-module ()
  "Select Prolog module name appropriate for current buffer.
Bases decision on buffer contents (-*- line)."
  ;; Look for -*- ... module: MODULENAME; ... -*-
  (let (beg end)
    (save-excursion
      (goto-char (point-min))
      (skip-chars-forward " \t")
      (and (search-forward "-*-" (save-excursion (end-of-line) (point)) t)
	   (progn
	     (skip-chars-forward " \t")
	     (setq beg (point))
	     (search-forward "-*-" (save-excursion (end-of-line) (point)) t))
	   (progn
	     (forward-char -3)
	     (skip-chars-backward " \t")
	     (setq end (point))
	     (goto-char beg)
	     (and (let ((case-fold-search t))
		    (search-forward "module:" end t))
		  (progn
		    (skip-chars-forward " \t")
		    (setq beg (point))
		    (if (search-forward ";" end t)
			(forward-char -1)
		      (goto-char end))
		    (skip-chars-backward " \t")
		    (buffer-substring beg (point)))))))))

(defun prolog-build-prolog-command (compilep file buffername)
  "Make Prolog command for FILE compilation/consulting.
If COMPILEP is non-nil, consider compilation, otherwise consulting."
  (let* ((compile-string
	  (if compilep prolog-compile-string-i prolog-consult-string-i))
	 (module (prolog-buffer-module))
	 (file-name (concat "'" file "'"))
	 (module-name (if module (concat "'" module "'")))
	 (module-file (if module
			  (concat module-name ":" file-name)
			file-name))
	 strbeg strend)
    (if (not (string-match "^'.*'$" buffername)) ; Add quotes
	(setq buffername (concat "'" buffername "'")))
    (while (string-match "%m" compile-string)
      (setq strbeg (substring compile-string 0 (match-beginning 0)))
      (setq strend (substring compile-string (match-end 0)))
      (setq compile-string (concat strbeg module-file strend)))
    (while (string-match "%f" compile-string)
      (setq strbeg (substring compile-string 0 (match-beginning 0)))
      (setq strend (substring compile-string (match-end 0)))
      (setq compile-string (concat strbeg file-name strend)))
    (while (string-match "%b" compile-string)
      (setq strbeg (substring compile-string 0 (match-beginning 0)))
      (setq strend (substring compile-string (match-end 0)))
      (setq compile-string (concat strbeg buffername strend)))
    (concat compile-string "\n")))

;;; The rest of this page is experimental code!

;; Global variables for process filter function
(defvar prolog-process-flag nil
  "Non-nil means that a prolog task (i.e. a consultation or compilation job) 
is running.")
(defvar prolog-consult-compile-output ""
  "Holds the unprocessed output from the current prolog task.")
(defvar prolog-consult-compile-first-line 1
  "The number of the first line of the file to consult/compile. Used for
temporary files.")
(defvar prolog-consult-compile-file nil
  "The file to compile/consult (can be a temporary file).")
(defvar prolog-consult-compile-real-file nil
  "The file name of the buffer to compile/consult.")

(defun prolog-consult-compile (compilep file &optional first-line)
  "Consult/compile FILE.
If COMPILEP is non-nil, perform compilation, otherwise perform CONSULTING.
COMMAND is a string described by the variables `prolog-consult-string'
and `prolog-compile-string'.
Optional argument FIRST-LINE is the number of the first line in the compiled
region.

This function must be called from the source code buffer."
  (if prolog-process-flag
      (error "Another Prolog task is running."))
  (prolog-ensure-process t)
  (let* ((buffer (get-buffer-create prolog-compilation-buffer))
	 (real-file buffer-file-name)
	 (command-string (prolog-build-prolog-command compilep file real-file))
	 (process (get-process "prolog"))
	 (old-filter (process-filter process)))
    (save-excursion
      (set-buffer buffer)
      (delete-region (point-min) (point-max))
      (compilation-mode)
      ;; Setting up font-locking for this buffer
      (make-local-variable 'font-lock-defaults)
      (setq font-lock-defaults 
	    '(prolog-font-lock-keywords nil nil ((?_ . "w"))))
      (if (eq prolog-system 'sicstus)
	  (progn
	    (make-local-variable 'compilation-parse-errors-function)
	    (setq compilation-parse-errors-function
		  'prolog-parse-sicstus-compilation-errors)))
      (insert command-string "\n"))
    (save-selected-window
      (pop-to-buffer buffer))
    (setq prolog-process-flag t
	  prolog-consult-compile-output ""
	  prolog-consult-compile-first-line (if first-line (1- first-line) 0)
	  prolog-consult-compile-file file
	  prolog-consult-compile-real-file (if (string= 
						file buffer-file-name)
					       nil
					     real-file))
    (save-excursion
      (set-buffer buffer)
      (goto-char (point-max))
      (set-process-filter process 'prolog-consult-compile-filter)
      (process-send-string
       "prolog"
       (prolog-build-prolog-command compilep file real-file))
      (while (and prolog-process-flag
		  (accept-process-output process 10)) ; 10 secs is ok?
	(sit-for 0.1)
	(if (not (get-process "prolog"))
	    (setq prolog-process-flag nil)))
      (insert (if compilep
		  "\nCompilation finished.\n"
		"\nConsulted.\n"))
      (set-process-filter process old-filter))))

(defun prolog-parse-sicstus-compilation-errors (limit atleast)
  "Parse the prolog compilation buffer for errors.
For use with the `compilation-parse-errors-function' variable."
  (setq compilation-error-list nil)
  (message "Parsing SICStus error messages...")
  (let (filepath dir file errorline)
    (while 
	(re-search-backward
	 "{\\([a-zA-Z ]* ERROR\\|Warning\\):.* in line[s ]*\\([0-9]+\\)"
	 limit t)
      (setq errorline (string-to-number (match-string 2)))
      (save-excursion
	(re-search-backward
	 "{\\(consulting\\|compiling\\|processing\\) \\(.*\\)\\.\\.\\.}"
	 limit t)
	(setq filepath (match-string 2)))

;;; Does this work with SICStus under Windows (i.e. backslahes and stuff?)
      (if (string-match "\\(.*/\\)\\([^/]*\\)$" filepath)
	  (progn
	    (setq dir (match-string 1 filepath))
	    (setq file (match-string 2 filepath))))

      (setq compilation-error-list
	    (cons
	     (cons (save-excursion
		     (beginning-of-line)
		     (point-marker))
		   (list (list file dir) errorline))
	     compilation-error-list)
	    ))
    ))

(defun prolog-consult-compile-filter (process output)
  "Filter function for Prolog compilation process."
  ;;(message "start")
  (setq prolog-consult-compile-output
	(concat prolog-consult-compile-output output))
  ;;(message "pccf1: %s" prolog-consult-compile-output)
  ;; Iterate through the lines of prolog-consult-compile-output
  (let (outputtype)
    (while (and prolog-process-flag
		(or
		 ;; Trace question
		 (progn 
		   (setq outputtype 'trace)
		   (and (eq prolog-system 'sicstus)
			(string-match
			 "^[ \t]*[0-9]+[ \t]*[0-9]+[ \t]*Call:.*? "
			 prolog-consult-compile-output)))
		 
		 ;; Match anything
		 (progn 
		   (setq outputtype 'normal)
		   (string-match "^.*\n" prolog-consult-compile-output))
		   ))
      ;;(message "outputtype: %s" outputtype)

      (setq output (match-string 0 prolog-consult-compile-output))
      ;; remove the text in output from prolog-consult-compile-output
      (setq prolog-consult-compile-output
	    (substring prolog-consult-compile-output (length output)))
      ;;(message "pccf2: %s" prolog-consult-compile-output)
      
      ;; If temporary files were used, then we change the error
      ;; messages to point to the original source file.
      (cond

       ;; If the prolog process was in trace mode then it requires
       ;; user input
       ((and (eq prolog-system 'sicstus) 
	     (eq outputtype 'trace))
	(let (input)
	  (setq input (concat (read-string output) "\n"))
	  (process-send-string "prolog" input)
	  (setq output (concat output input))))

       ((eq prolog-system 'sicstus)
	(if (and prolog-consult-compile-real-file
		 (string-match
		  "\\({.*:.* in line[s ]*\\)\\([0-9]+\\)-\\([0-9]+\\)" output))
	    (setq output (replace-match
			  ;; Adds a {processing ...} line so that 
			  ;; `prolog-parse-sicstus-compilation-errors'
			  ;; finds the real file instead of the temporary one.
			  ;; Also fixes the line numbers.
			  (format "Added by Emacs: {processing %s...}\n%s%d-%d"
				  prolog-consult-compile-real-file
				  (match-string 1 output)
				  (+ prolog-consult-compile-first-line
				     (string-to-number
				      (match-string 2 output)))
				  (+ prolog-consult-compile-first-line
				     (string-to-number
				      (match-string 3 output))))
			  t t output)))
	)
       
       ((eq prolog-system 'swi)
	(if (and prolog-consult-compile-real-file
		 (string-match (format
				"%s\\([ \t]*:[ \t]*\\)\\([0-9]+\\)"
				prolog-consult-compile-file)
			       output))
	    (setq output (replace-match
			  ;; Real filename + text + fixed linenum
			  (format "%s%s%d"
				  prolog-consult-compile-real-file
				  (match-string 1 output)
				  (+ prolog-consult-compile-first-line
				     (string-to-number
				      (match-string 2 output))))
			  t t output)))
	)
       
       (t ())
       )
      ;; Write the output in the *prolog-compilation* buffer
      (insert output)))

  ;; If the prompt is visible, then the task is finished
  (if (string-match prolog-prompt-regexp-i prolog-consult-compile-output)
      (setq prolog-process-flag nil)))

(defun prolog-consult-compile-file (compilep)
  "Consult/compile file of current buffer.
If COMPILEP is non-nil, compile, otherwise consult."
  (let ((file buffer-file-name))
    (if file
	(progn
	  (save-some-buffers)
	  (prolog-consult-compile compilep file))
      (prolog-consult-compile-region compilep (point-min) (point-max)))))

(defun prolog-consult-compile-buffer (compilep)
  "Consult/compile current buffer.
If COMPILEP is non-nil, compile, otherwise consult."
  (prolog-consult-compile-region compilep (point-min) (point-max)))

(defun prolog-consult-compile-region (compilep beg end)
  "Consult/compile region.
If COMPILEP is non-nil, compile, otherwise consult."
  (let ((file prolog-temp-filename)
	(lines (count-lines 1 beg)))
    (write-region beg end file nil 'no-message)
    (write-region "\n" nil file t 'no-message)
    (prolog-consult-compile compilep file
			    (if (looking-at "^") (1+ lines) lines))
    (delete-file file)))

(defun prolog-consult-compile-predicate (compilep)
  "Consult/compile the predicate around current point.
If COMPILEP is non-nil, compile, otherwise consult."
  (prolog-consult-compile-region
   compilep (prolog-pred-start) (prolog-pred-end)))


;;-------------------------------------------------------------------
;; Font-lock stuff
;;-------------------------------------------------------------------

;; Auxilliary functions
(defun prolog-make-keywords-regexp (keywords &optional protect)
  "Create regexp from the list of strings.
If PROTECT is non-nil, surround the result regexp by word breaks."
  (let ((regexp
	 (if (fboundp 'regexp-opt)
	     ;; Emacs 20
	     (regexp-opt keywords)
	   ;; Older Emacsen
	   (concat (mapconcat 'identity keywords  "\\|")))
	 ))
    (if protect
	(concat "\\<\\(" regexp "\\)\\>")
      regexp)))

(defun prolog-font-lock-object-matcher (bound)
  "Find SICStus objects method name for font lock."
  (let (point
	 (case-fold-search nil))
    (while (and (not point)
		(re-search-forward "\\(::[ \t\n]*{\\|&\\)[ \t]*"
				   bound t))
      (while (or (re-search-forward "\\=\n[ \t]*" bound t)
		  (re-search-forward "\\=%.*" bound t)
		  (and (re-search-forward "\\=/\\*" bound t)
		       (re-search-forward "\\*/[ \t]*" bound t))))
      (setq point (re-search-forward
		   (format "\\=\\(%s\\)" prolog-atom-regexp)
		   bound t)))
    point))

;; Set everything up
(defun prolog-font-lock-keywords ()
  "Set up font lock keywords for the current Prolog system."
  (when window-system
    (require 'font-lock)
    
    ;; Define Prolog faces
    (setq prolog-warning-face 'font-lock-warning-face
	  prolog-builtin-face 'font-lock-builtin-face)
    (defface prolog-redo-face
      '((((class grayscale)) (:italic t))
	(((class color)) (:foreground "darkorchid"))
	(t (:italic t)))
      "Prolog mode face for highlighting redo trace lines."
      :group 'prolog-faces)
    (defface prolog-exit-face
      '((((class grayscale)) (:underline t))
	(((class color) (background dark)) (:foreground "green"))
	(((class color) (background light)) (:foreground "ForestGreen"))
	(t (:underline t)))
      "Prolog mode face for highlighting exit trace lines."
      :group 'prolog-faces)
    (defface prolog-exception-face
      '((((class grayscale)) (:bold t :italic t :underline t))
	(((class color)) (:bold t :foreground "black" :background "Khaki"))
	(t (:bold t :italic t :underline t)))
      "Prolog mode face for highlighting exception trace lines."
      :group 'prolog-faces)
  
    ;; Font Lock Patterns
    (let (
	  ;; "Native" Prolog patterns
	  (head-predicates
	   (list (format "^%s" prolog-atom-regexp)
		 0 font-lock-function-name-face))
	  (variables
	   '("\\<\\([_A-Z][a-zA-Z0-9_]*\\)"
	     1 font-lock-variable-name-face))
	  (important-elements
	   (list (if (eq prolog-system 'mercury)
		     "[][}{;|]\\|\\\\[+=]\\|<?=>?"
		   "[][}{!;|]\\|\\*->")
		 0 'font-lock-keyword-face))
	  (important-elements-1
	   '("[^-*]\\(->\\)" 1 font-lock-keyword-face))
	  (predspecs			; module:predicate/cardinality
	   (list (format "\\<\\(%s:\\|\\)%s/[0-9]+"
			 prolog-atom-regexp prolog-atom-regexp)
		 0 font-lock-function-name-face 'prepend))
	  (keywords			; directives (queries)
	   (list
	    (if (eq prolog-system 'mercury)
		(concat
		 "\\<\\("
		 (prolog-make-keywords-regexp prolog-keywords-i)
		 "\\|"
		 (prolog-make-keywords-regexp
		  prolog-determinism-specificators-i)
		 "\\)\\>")
	      (concat
	       "^[?:]- *\\("
	       (prolog-make-keywords-regexp prolog-keywords-i)
	       "\\)\\>"))
	      1 'prolog-builtin-face))
	  (quoted_atom (list prolog-quoted-atom-regexp
			     2 'font-lock-string-face 'append))
	  (string (list prolog-string-regexp
			1 'font-lock-string-face 'append))
	  ;; SICStus specific patterns
	  (sicstus-object-methods
	   (if (eq prolog-system 'sicstus)
	       '(prolog-font-lock-object-matcher
		 1 font-lock-function-name-face)))
	  ;; Mercury specific patterns
	  (types
	   (if (eq prolog-system 'mercury)
	       (list
		(prolog-make-keywords-regexp prolog-types-i t)
		0 'font-lock-type-face)))
	  (modes
	   (if (eq prolog-system 'mercury)
	       (list
		(prolog-make-keywords-regexp prolog-mode-specificators-i t)
		0 'font-lock-reference-face)))
	  (directives
	   (if (eq prolog-system 'mercury)
	       (list
		(prolog-make-keywords-regexp prolog-directives-i t)
		0 'prolog-warning-face)))
	  ;; Inferior mode specific patterns
	  (prompt
	   (list prolog-prompt-regexp-i 0 'font-lock-keyword-face))
	  (trace-exit
	   (cond
	    ((eq prolog-system 'sicstus)
	     '("[ \t]*[0-9]+[ \t]+[0-9]+[ \t]*\\(Exit\\):"
	       1 'prolog-exit-face))
	    ((eq prolog-system 'swi)
	     '("[ \t]*\\(Exit\\):[ \t]*([ \t0-9]*)" 1 prolog-exit-face))
	    (t nil)))
	  (trace-fail
	   (cond
	    ((eq prolog-system 'sicstus)
	     '("[ \t]*[0-9]+[ \t]+[0-9]+[ \t]*\\(Fail\\):"
	       1 prolog-warning-face))
	    ((eq prolog-system 'swi)
	     '("[ \t]*\\(Fail\\):[ \t]*([ \t0-9]*)" 1 prolog-warning-face))
	    (t nil)))
	  (trace-redo
	   (cond
	    ((eq prolog-system 'sicstus)
	     '("[ \t]*[0-9]+[ \t]+[0-9]+[ \t]*\\(Redo\\):"
	       1 'prolog-redo-face))
	    ((eq prolog-system 'swi)
	     '("[ \t]*\\(Redo\\):[ \t]*([ \t0-9]*)" 1 prolog-redo-face))
	    (t nil)))
	  (trace-call
	   (cond
	    ((eq prolog-system 'sicstus)
	     '("[ \t]*[0-9]+[ \t]+[0-9]+[ \t]*\\(Call\\):"
	       1 font-lock-function-name-face))
	    ((eq prolog-system 'swi)
	     '("[ \t]*\\(Call\\):[ \t]*([ \t0-9]*)"
	       1 font-lock-function-name-face))
	    (t nil)))
	  (trace-exception
	   (cond
	    ((eq prolog-system 'sicstus)
	     '("[ \t]*[0-9]+[ \t]+[0-9]+[ \t]*\\(Exception\\):"
	       1 prolog-exception-face))
	    ((eq prolog-system 'swi)
	     '("[ \t]*\\(Exception\\):[ \t]*([ \t0-9]*)"
	       1 prolog-exception-face))
	    (t nil)))
	  (error-message-identifier
	   (cond
	    ((eq prolog-system 'sicstus)
	     '("{\\([A-Z]* ?ERROR:\\)" 1 prolog-exception-face prepend))
	    ((eq prolog-system 'swi)
	     '("^[[]\\(WARNING:\\)" 1 prolog-builtin-face prepend))
	    (t nil)))
	  (error-whole-messages
	   (cond
	    ((eq prolog-system 'sicstus)
	     '("{\\([A-Z]* ?ERROR:.*\\)}[ \t]*$"
	       1 font-lock-comment-face append))
	    ((eq prolog-system 'swi)
	     '("^[[]WARNING:[^]]*[]]$" 0 font-lock-comment-face append))
	    (t nil)))
	  (error-warning-messages
	   ;; Mostly errors that SICStus asks the user about how to solve,
	   ;; such as "NAME CLASH:" for example.
	   (cond
	    ((eq prolog-system 'sicstus)
	     '("^[A-Z ]*[A-Z]+:" 0 prolog-warning-face))
	    (t nil)))
	  (warning-messages
	   (cond
	    ((eq prolog-system 'sicstus)
	     '("\\({ ?\\(Warning\\|WARNING\\) ?:.*}\\)[ \t]*$" 
	       2 prolog-warning-face prepend))
	    (t nil))))

      ;; Make font lock list
      (delq
       nil
       (cond
	((eq major-mode 'prolog-mode)
	 (list
	  head-predicates
	  quoted_atom
	  string
	  variables
	  important-elements
	  important-elements-1
	  predspecs
	  keywords
	  sicstus-object-methods
	  types
	  modes
	  directives))
	((eq major-mode 'prolog-inferior-mode)
	 (list
	 prompt
	 error-message-identifier
	 error-whole-messages
	 error-warning-messages
	 warning-messages
	 predspecs
	 trace-exit
	 trace-fail
	 trace-redo
	 trace-call
	 trace-exception))
	((eq major-mode 'compilation-mode)
	 (list
	 error-message-identifier
	 error-whole-messages
	 error-warning-messages
	 warning-messages
	 predspecs))))
      )))


;;-------------------------------------------------------------------
;; Indentation stuff
;;-------------------------------------------------------------------

(defun prolog-indent-line (&optional whole-exp)
  "Indent current line as Prolog code.
With argument, indent any additional lines of the same clause
rigidly along with this one (not yet)."
  (interactive "p")
  (let ((indent (prolog-indent-level))
	(pos (- (point-max) (point))) beg)
    (beginning-of-line)
    (setq beg (point))
    (skip-chars-forward " \t")
    (if (zerop (- indent (current-column)))
	nil
      (delete-region beg (point))
      (indent-to indent))
    (if (> (- (point-max) pos) (point))
	(goto-char (- (point-max) pos)))
    
    ;; Align comments
    (if prolog-align-comments-flag
	(save-excursion
	  (prolog-goto-comment-column t)))

    (prolog-insert-spaces-after-paren)

    ))

(defun prolog-comment-indent ()
  "Compute prolog comment indentation."
  (cond ((looking-at "%%%") (prolog-indentation-level-of-line))
	((looking-at "%%") (prolog-indent-level))
	(t
	 (save-excursion
	   (skip-chars-backward " \t")
	   ;; Insert one space at least, except at left margin.
	   (max (+ (current-column) (if (bolp) 0 1))
		comment-column)))
	))

(defun prolog-indent-level ()
  "Compute prolog indentation level."
  (save-excursion
    (beginning-of-line)
    (let ((totbal (prolog-region-paren-balance
		   (prolog-clause-start t) (point)))
	  (oldpoint (point)))
      (skip-chars-forward " \t")
      (cond
       ((looking-at "%%%") (prolog-indentation-level-of-line))
					;Large comment starts
       ((looking-at "%[^%]") comment-column) ;Small comment starts
       ((bobp) 0)			;Beginning of buffer

       ;; If we found '}' then we must check if it's the
       ;; end of an object declaration or something else.
       ((and (looking-at "}") 
	     (save-excursion
	       (forward-char 1)
	       ;; Goto to matching {
	       (if prolog-use-prolog-tokenizer-flag
		   (prolog-backward-list)
		 (backward-list))
	       (skip-chars-backward " \t")
	       (backward-char 2)
	       (looking-at "::")))
	;; It was an object
	(if prolog-object-end-to-0-flag
	    0
	  prolog-indent-width)) 

       ;;End of /* */ comment
       ((looking-at "\\*/")                  
	(save-excursion
	  (prolog-find-start-of-mline-comment)
	  (skip-chars-backward " \t")
	  (- (current-column) 2)))

       ;; Here we check if the current line is within a /* */ pair
       ((and (looking-at "[^%/]")
	     (eq (prolog-in-string-or-comment) 'cmt)) 
	(if prolog-indent-mline-comments-flag
	    (prolog-find-start-of-mline-comment)
	  ;; Same as before
	  (prolog-indentation-level-of-line)))

       (t
	(let ((empty t) ind linebal)
	  ;; See previous indentation
	  (while empty
	    (forward-line -1)
	    (beginning-of-line)
	    (if (= (point) (point-min))
		(setq empty nil)
	      (skip-chars-forward " \t")
	      (if (not (or (prolog-in-string-or-comment)
			   (looking-at "%") 
			   (looking-at "\n")))
		  (setq empty nil))))

	  ;; Store this line's indentation
	  (if (= (point) (point-min))
	      (setq ind 0)		;Beginning of buffer
	    (setq ind (current-column))) ;Beginning of clause

	  ;; Compute the balance of the line
	  (setq linebal (prolog-paren-balance))

	  (if (< linebal 0)
	      (progn
		;; Add 'indent-level' mode to find-unmatched-paren instead?
		(end-of-line)
		(setq ind (prolog-find-indent-of-matching-paren))))

	  (beginning-of-line)

	  ;; Check if the line ends with ":-", ".", ":: {", "}" (might be
	  ;; unnecessary), "&" or ")" (The last four concerns SICStus objects)
	  (cond
	   ;; If the last char of the line is a '&' then set the indent level
	   ;; to prolog-indent-width (used in SICStus objects)
	   ((and (eq prolog-system 'sicstus) 
		 (looking-at ".+&[ \t]*\\(%.*\\|\\)$"))
	    (setq ind prolog-indent-width))

	   ;; Increase indentation if the previous line was the head of a rule
	   ;; and does not contain a '.'
	   ((and (looking-at (format ".*%s[^\\.]*[ \t]*\\(%%.*\\|\\)$" 
				     prolog-head-delimiter))
		 ;; We must check that the match is at a paren balance of 0.
		 (save-excursion
		   (let ((p (point)))
		     (re-search-forward prolog-head-delimiter)
		     (>= 0 (prolog-region-paren-balance p (point))))))
	    (let (headindent)
	      (if (< (prolog-paren-balance) 0)
		  (save-excursion
		    (end-of-line)
		    (setq headindent (prolog-find-indent-of-matching-paren)))
		(setq headindent (prolog-indentation-level-of-line)))
	      (setq ind (+ headindent prolog-indent-width))))

	   ;; The previous line was the head of an object
	   ((looking-at ".+ *::.*{[ \t]*$")
	    (setq ind prolog-indent-width))

	   ;; If a '.' is found at the end of the previous line, then decrease
	   ;; the indentation. (The \\(%.*\\|\\) part of the regexp is for
	   ;; comments at the end of the line)
	   ((and (looking-at "^.+\\.[ \t]*\\(%.*\\|\\)$") 
		 ;; Make sure that the '.' found is not in a comment or string
		 (save-excursion
		   (end-of-line)
		   (re-search-backward "\\.[ \t]*\\(%.*\\|\\)$" (point-min))
		   (not (prolog-in-string-or-comment))))
	    (setq ind 0))

	   )

	  ;; If the last non comment char is a ',' or left paren or a left-
	  ;; indent-regexp then indent to open parenthesis level
	  (if (and
	       (> totbal 0)
	       ;; SICStus objects have special syntax rules if point is
	       ;; not inside additional parens (objects are defined
	       ;; within {...})
	       (not (and (eq prolog-system 'sicstus)
			 (= totbal 1)
			 (prolog-in-object))))
	      (if (looking-at
		   (format "\\(%s\\|%s\\|0'.\\|[0-9]+'[0-9a-zA-Z]+\\|[^\n\'\"%%]\\)*\\(,\\|%s\\|%s\\)\[ \t]*\\(%%.*\\|\\)$" 
			   prolog-quoted-atom-regexp prolog-string-regexp
			   prolog-left-paren prolog-left-indent-regexp))
		  (progn
		    (goto-char oldpoint)
		    (setq ind (prolog-find-unmatched-paren 'termdependent)))
		(goto-char oldpoint)
		(setq ind (prolog-find-unmatched-paren nil))
		))
	  

	  ;; Return the indentation level
	  ind
	  ))))))

(defun prolog-find-indent-of-matching-paren ()
  "Find the indentation level based on the matching parenthesis.
Indentation level is set to the one the point is after when the function is
called."
  (save-excursion
    ;; Go to the matching paren
    (if prolog-use-prolog-tokenizer-flag
	(prolog-backward-list)
      (backward-list))

    ;; If this was the first paren on the line then return this line's
    ;; indentation level
    (if (prolog-paren-is-the-first-on-line-p)
	(prolog-indentation-level-of-line)
      ;; It was not the first one
      (progn
	 ;; Find the next paren
	 (prolog-goto-next-paren 0)

	 ;; If this paren is a left one then use its column as indent level,
	 ;; if not then recurse this function
	 (if (looking-at prolog-left-paren)
	     (+ (current-column) 1)
	   (progn
	      (forward-char 1)
	      (prolog-find-indent-of-matching-paren)))
	 ))
    ))

(defun prolog-indentation-level-of-line ()
  "Return the indentation level of the current line."
  (save-excursion
    (beginning-of-line)
    (skip-chars-forward " \t")
    (current-column)))

(defun prolog-first-pos-on-line ()
  "Return the first position on the current line."
  (save-excursion
    (beginning-of-line)
    (point)))

(defun prolog-paren-is-the-first-on-line-p ()
  "Return t if the parenthesis under the point is the first one on the line.
Return nil otherwise.
Note: does not check if the point is actually at a parenthesis!"
  (save-excursion
    (let ((begofline (prolog-first-pos-on-line)))
      (if (= begofline (point))
	  t
	(if (prolog-goto-next-paren begofline)
	    nil
	  t)))))

(defun prolog-find-unmatched-paren (&optional mode)
  "Return the column of the last unmatched left parenthesis.
If MODE is `skipwhite' then any white space after the parenthesis is added to
the answer.
If MODE is `plusone' then the parenthesis' column +1 is returned.
If MODE is `termdependent' then if the unmatched parenthesis is part of
a compound term the function will work as `skipwhite', otherwise
it will return the column paren plus the value of `prolog-paren-indent'.
If MODE is nil or not set then the parenthesis' exact column is returned."
  (save-excursion
    ;; If the next paren we find is a left one we're finished, if it's
    ;; a right one then we go back one step and recurse
    (prolog-goto-next-paren 0)

    (let ((roundparen (looking-at "(")))
      (if (looking-at prolog-left-paren)
	  (let ((notpartofterm 
		 (save-excursion
		   (backward-char 1)
		   (looking-at "[ \t]"))))
	    (if (eq mode nil)
		(current-column)
	      (if (and roundparen
		       (eq mode 'termdependent) 
		       notpartofterm)
		  (+ (current-column) prolog-paren-indent)
		(forward-char 1)
		(if (or (eq mode 'skipwhite) (eq mode 'termdependent) )
		    (skip-chars-forward " \t"))
		(current-column))))
	;; Not looking at left paren
	(progn
	  (forward-char 1)
	  ;; Go to the matching paren. When we get there we have a total
	  ;; balance of 0.
	  (if prolog-use-prolog-tokenizer-flag
	      (prolog-backward-list)
	    (backward-list))
	  (prolog-find-unmatched-paren mode)))
      )))


(defun prolog-paren-balance ()
  "Return the parenthesis balance of the current line.
A return value of n means n more left parentheses than right ones."
  (save-excursion
    (end-of-line)
    (prolog-region-paren-balance (prolog-first-pos-on-line) (point))))

(defun prolog-region-paren-balance (start end)
  "Return the summed parenthesis balance in the region.
The region is limited by START and END positions."
  (save-excursion
    (let ((state (if prolog-use-prolog-tokenizer-flag
		     (prolog-tokenize start end)
		   (parse-partial-sexp start end))))
      (nth 0 state))))

(defun prolog-goto-next-paren (limit-pos)
  "Move the point to the next parenthesis earlier in the buffer.
Return t if a match was found before LIMIT-POS.  Return nil otherwise."
  (let (retval)
    (setq retval (re-search-backward
		  (concat prolog-left-paren "\\|" prolog-right-paren)
		  limit-pos t))

    ;; If a match was found but it was in a string or comment, then recurse
    (if (and retval (prolog-in-string-or-comment))
	(prolog-goto-next-paren limit-pos)
      retval)
    ))

(defun prolog-in-string-or-comment ()
  "Check whether string, atom, or comment is under current point.
Return:
 `txt' if the point is in a string, atom, or character code expression
 `cmt' if the point is in a comment
 nil otherwise."
  (save-excursion
    (let* ((start
	    (if (eq prolog-parse-mode 'beg-of-line)
		;; 'beg-of-line
		(save-excursion
		  (let (safepoint)
		    (beginning-of-line)
		    (setq safepoint (point))
		    (while (and (> (point) (point-min))
				(progn
				  (forward-line -1)
				  (end-of-line)
				  (if (not (bobp))
				      (backward-char 1))
				  (looking-at "\\\\"))
				)
		      (beginning-of-line)
		      (setq safepoint (point)))
		    safepoint))
	      ;; 'beg-of-clause
	      (prolog-clause-start)))
	   (end (point))
	   (state (if prolog-use-prolog-tokenizer-flag
		      (prolog-tokenize start end)
		    (parse-partial-sexp start end))))
      (cond
       ((nth 3 state) 'txt) ; String
       ((nth 4 state) 'cmt) ; Comment
       (t
	(cond
	 ((looking-at "%") 'cmt) ; Start of a comment
	 ((looking-at "/\\*") 'cmt) ; Start of a comment
	 ;;((looking-at "\'") 'txt) ; Start of an atom
	 ;;((looking-at "\"") 'txt) ; Start of a string
	 (t nil)
	 ))))
    ))

(defun prolog-find-start-of-mline-comment ()
  "Return the start column of a /* */ comment.
This assumes that the point is inside a comment."
  (re-search-backward "/\\*" (point-min) t)
  (forward-char 2)
  (skip-chars-forward " \t")
  (current-column))

(defun prolog-insert-spaces-after-paren ()
  "Insert spaces after the opening parenthesis.
Spaces are inserted if the parenthesis is the first non-white character
on the line."
  (save-excursion
    (let (incr level)
      (beginning-of-line)
      (skip-chars-forward " \t")
      (if (looking-at (format "(\\|%s" prolog-left-indent-regexp))
	  (progn
	    ;; Treat "( If -> " lines specially.
	    (if (looking-at "(.*->")
		(setq incr 2)
	      (setq incr prolog-paren-indent))
	    
	    ;; Goto the first char after the matched text
	    (re-search-forward (format "(\\|%s" prolog-left-indent-regexp)
			       (point-max) t)
	    (goto-char (match-end 0))
	    
	    (setq level (+ (prolog-find-unmatched-paren) incr))
	    
	    ;; Remove old white space
	    (let ((start (point)))
	      (skip-chars-forward " \t")
	      (delete-region start (point)))
	    
	    (indent-to level)
	    ))))
  (skip-chars-forward " \t"))


;;-------------------------------------------------------------------
;; The tokenizer
;;-------------------------------------------------------------------

(defconst prolog-tokenize-searchkey
  (concat "[0-9]+'"
	  "\\|"
	  "['\"]"
	  "\\|"
	  prolog-left-paren
	  "\\|"
	  prolog-right-paren
	  "\\|"
	  "%"
	  "\\|"
	  "/\\*"
	  ))

(defun prolog-tokenize ( start end &optional stopcond )
  "Tokenize a region of prolog code.
START and END are positions.
STOPCOND decides the stop condition of the parsing. Valid values
are 'zerodepth which stops the parsing at the first right parenthesis
where the parenthesis depth is zero, 'skipover which skips over
the current entity (e.g. a list, a string, etc.) and nil.

The function returns a list with the following information:
 0. parenthesis depth 
 3. 'atm if END is inside an atom
    'str if END is inside a string
    'chr if END is in a character code expression (0'x)
    nil otherwise
 4. non-nil if END is inside a comment
 5. end position (always equal to END if STOPCOND is nil)
The rest of the elements are undefined."
  (save-excursion
    (let* ((end2 (1+ end))
	   oldp
	   (depth 0)
	   (quoted nil)
	   inside_cmt
	   (endpos end2)
	   skiptype ; The type of entity we'll skip over
	   )
      (goto-char start)

      (if (and (eq stopcond 'skipover)
	       (looking-at "[^[({'\"]"))
	  (setq endpos (point))		; Stay where we are
	(while (and
		(re-search-forward prolog-tokenize-searchkey end2 t)
		(< (point) end2))
	  (progn
	    (setq oldp (point))
	    (goto-char (match-beginning 0))
	    (cond
	     ;; Atoms and strings
	     ((looking-at "'")
	      ;; Find end of atom
	      (if (re-search-forward "[^\\]'" end2 'limit)
		  ;; Found end of atom
		  (progn
		    (setq oldp end2)
		    (if (and (eq stopcond 'skipover)
			     (not skiptype))
			(setq endpos (point))
		      (setq oldp (point)))) ; Continue tokenizing
		(setq quoted 'atm)))
	   
	     ((looking-at "\"")
	      ;; Find end of string
	      (if (re-search-forward "[^\\]\"" end2 'limit)
		  ;; Found end of string
		  (progn
		    (setq oldp end2)
		    (if (and (eq stopcond 'skipover)
			     (not skiptype))
			(setq endpos (point))
		      (setq oldp (point)))) ; Continue tokenizing
		(setq quoted 'str)))

	     ;; Paren stuff
	     ((looking-at prolog-left-paren)
	      (setq depth (1+ depth))
	      (setq skiptype 'paren))

	     ((looking-at prolog-right-paren)
	      (setq depth (1- depth))
	      (if (and
		   (or (eq stopcond 'zerodepth)
		       (and (eq stopcond 'skipover) 
			    (eq skiptype 'paren)))
		   (= depth 0))
		  (progn
		    (setq endpos (1+ (point)))
		    (setq oldp end2))))

	     ;; Comment stuff
	     ((looking-at comment-start)
	      (end-of-line)
	      (if (>= (point) end)
;	      (if (>= (point) end2)
		  (progn
		    (setq inside_cmt t)
		    (setq oldp end2))
		(setq oldp (point))))

	     ((looking-at "/\\*")
	      (if (re-search-forward "\\*/" end2 'limit)
		  (setq oldp (point))
		(setq inside_cmt t)
		(setq oldp end2)))

	     ;; 0'char
	     ((looking-at "0'")
	      (setq oldp (1+ (match-end 0)))
	      (if (> oldp end) 
		  (setq quoted 'chr)))
	   
	     ;; base'number
	     ((looking-at "[0-9]+'")
	      (goto-char (match-end 0))
	      (skip-chars-forward "0-9a-zA-Z")
	      (setq oldp (point)))

	     
	     )
	    (goto-char oldp)
	    ))				; End of while
	)

      ;; Create return list
      (list depth nil nil quoted inside_cmt endpos)
      )))


;;-------------------------------------------------------------------
;; Online help
;;-------------------------------------------------------------------

(defvar prolog-help-function
  '((mercury nil)
    (sicstus prolog-help-info)
    (swi prolog-help-online)
    (t prolog-help-online))
  "Alist for the name of the function which finds help on a predicate.")

(defun prolog-help-on-predicate ()
  "Invoke online help on the atom under cursor."
  (interactive)
  (let* (word
	 predicate
	 point)
    (setq word (prolog-atom-under-point))
    (setq predicate (read-from-minibuffer
		     (format "Help on predicate%s: "
			     (if word
				 (concat " (default " word ")")
			       ""))))
    (if (string= predicate "")
	(setq predicate word))
    (if prolog-help-function-i
	(funcall prolog-help-function-i predicate)
      (error "Sorry, no help method defined for this Prolog system."))))

(defun prolog-help-info (predicate)
  (let ((buffer (current-buffer))
	oldp
	(str (concat "^\\* " (regexp-quote predicate) " */")))
    (require 'info)
    (pop-to-buffer nil)
    (Info-goto-node prolog-info-predicate-index)
    (if (not (re-search-forward str nil t))
	(error (format "Help on predicate `%s' not found." predicate)))

    (setq oldp (point))
    (if (re-search-forward str nil t)
	;; Multiple matches, ask user
	(let (n (max 2))
	  ;; Count matches
	  (while (re-search-forward str nil t)
	    (setq max (1+ max)))

	  (goto-char oldp)
	  (re-search-backward "[^ /]" nil t)
	  (recenter 0)
	  (setq n (read-input 
		   (format "Several matches, choose (1-%d): " max) "1"))
	  (forward-line (- (string-to-number n) 1)))
      ;; Single match
      (re-search-backward "[^ /]" nil t))

    (Info-follow-nearest-node)
    (re-search-forward (concat "^`" (regexp-quote predicate)) nil t)
    (beginning-of-line)
    (recenter 0)
    (pop-to-buffer buffer)))

(defun prolog-help-online (predicate)
  (prolog-ensure-process)
  (process-send-string "prolog" (concat "help(" predicate ").\n"))
  (display-buffer "*prolog*"))

(defun prolog-help-apropos (string)
  "Find Prolog apropos on given STRING.
This function is only available when `prolog-system' is set to `swi'."
  (interactive "sApropos: ")
  (cond
   ((eq prolog-system 'swi)
    (prolog-ensure-process)
    (process-send-string "prolog" (concat "apropos(" string ").\n"))
    (display-buffer "*prolog*"))
   (t
    (error "Sorry, no Prolog apropos available for this Prolog system."))))

(defun prolog-atom-under-point ()
  "Return the atom under the point."
  (save-excursion
    (let* ((case-fold-search nil)
	   (regexp-1 prolog-atom-regexp)
	   (regexp-2 "[-<>=*+@$:\\]+")
	   (regexp (format "\\(\\<%s\\>\\|%s\\)" regexp-1 regexp-2))
	   (regexp-1-point (concat regexp-1 "\\="))
	   (regexp-2-point (concat regexp-2 "\\="))
	   atom)
      (if (or (looking-at prolog-atom-char-regexp)
	      (and (/= (point) (point-min)) ; Don't go past BOB
		   (save-excursion
		     (backward-char 1)
		     (looking-at prolog-atom-char-regexp))))
	  (re-search-backward "\\<" nil t))
      (setq point (if (re-search-forward regexp nil t)
		      (match-end 0)
		    (point)))
      (re-search-backward regexp nil t)
      (if (looking-at regexp-1)
	  (while (re-search-backward regexp-1-point nil t))
	(while (re-search-backward regexp-2-point nil t)))
      (setq atom (and (not (eq (point) point))
		      (buffer-substring-no-properties (point) point)))
      ;; Return atom
      atom)
    )
  )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Help function with completion
;; Stolen from Per Mildner's SICStus debugger mode and modified

(defun prolog-find-documentation ()
  "Go to the Info node for a predicate in the SICStus Info manual."
  (interactive)
  (let* ((pred (prolog-read-predicate)))
    (prolog-goto-predicate-info pred)))

;; Very similar to prolog-help-info except that that function cannot
;; cope with arity and that it asks the user if there are several
;; functors with different arity.
(defun prolog-goto-predicate-info (predicate)
  "Go to the info page for PREDICATE, which is a PredSpec."
  (interactive)
  (require 'info)
  (let ((buffer (current-buffer))
	;; Remove the arity from the predspec
	(name (progn
		(string-match "\\(.*\\)/[0-9]+$" predicate)
		(substring predicate 0 (match-end 1))))
	oldp
	(str (regexp-quote predicate)))
    (pop-to-buffer nil)

    (Info-goto-node prolog-info-predicate-index)
    (if (not (re-search-forward str nil t))
	(error (format "Help on predicate `%s' not found." predicate)))
    
    (setq oldp (point))
    (Info-follow-nearest-node)
    
    (re-search-forward (concat "^`" (regexp-quote name)) nil t)
    (beginning-of-line)
    (recenter 0)
    (pop-to-buffer buffer))
)

(defun prolog-read-predicate ()
  "Read a PredSpec from the user. The returned value is a 
string \"functor/arity\". The interaction supports completion."
  (let ((initial (prolog-atom-under-point))
	answer)
    ;; If the predicate index is not yet built, do it now 
    (if (not prolog-info-alist) 
	(prolog-build-info-alist))
    ;; Read the PredSpec from the user
    (setq answer (completing-read "Help on predicate: "
				  prolog-info-alist nil t initial))
    (if (equal answer "")
	initial
      answer)))

(defvar prolog-info-alist nil 
  "Alist with all builtin predicates. Only for internal use
by `prolog-find-documentation'")

(defun prolog-build-info-alist (&optional ignore) ; ignore is lazy flag
  "Alist of (\"name/arity\" . (Info-node1 Info-node2 ...)
for all builtins and library predicates. Typically there are just
 one Info-node associated with each name."
  (message "Building info alist...")      
  (setq prolog-info-alist (prolog-build-predicate-index-alist))
  (message "Building info alist... done."))

(defun prolog-build-predicate-index-alist ()
  "Build a list of (\"name/arity\" . (Node1 Node2...))
(usually just one node per name)"
  (let ((l '())
        (last-entry (cons "" nil)))
    (save-excursion
      (save-window-excursion
        ;; select any window but the minibuffer (as we cannot switch
        ;; buffers in minibuffer window.
        ;; I am not sure this is the right/best way
        (if (active-minibuffer-window)  ; nil if none active
            (select-window (next-window)))
        ;; Do this after going away from minibuffer window
        (save-window-excursion
          (info))
        (Info-goto-node prolog-info-predicate-index)
	(goto-char (point-min))
	(while (re-search-forward
		"^\\* \\(.+[a-z_A-Z0-0]*\\)/\\([0-9]+\\)" nil t)
	  (let* ((name (buffer-substring-no-properties
			(match-beginning 1) (match-end 1)))
		 (arity (string-to-int (buffer-substring-no-properties
					(match-beginning 2) (match-end 2))))
		 (fa (format "%s/%d" name arity))
		 (info-node (Info-extract-menu-node-name)))
	    (if (equal fa (car last-entry))
		(setcdr last-entry (cons info-node (cdr last-entry)))
	      (setq last-entry (cons fa (list info-node))
		    l (cons last-entry l)))))
	(nreverse l)
	))))


;;-------------------------------------------------------------------
;; Miscellaneous functions
;;-------------------------------------------------------------------

(defun prolog-temporary-file ()
  "Make temporary file name for compilation."
  (make-temp-name 
   (concat 
    (or
     (getenv "TEMP") 
     (getenv "TMP")
     (getenv "SYSTEMP")
     (getenv "TMPDIR")
     "/tmp")
    "/prolcomp")))
(setq prolog-temp-filename (prolog-temporary-file))

(defun prolog-goto-prolog-process-buffer ()
  "Switch to the prolog process buffer and go to its end."
  (switch-to-buffer-other-window "*prolog*")
  (goto-char (point-max))
)


(defun prolog-debug-on (&optional prefix)
  "Enable debugging.
When called with prefix argument, disable debugging instead."
  (interactive "P")
  (if prefix
      (prolog-debug-off)
    (prolog-process-insert-string (get-process "prolog")
				  prolog-debug-on-string)
    (process-send-string "prolog" prolog-debug-on-string)))

(defun prolog-debug-off ()
  "Disable debugging."
  (interactive)
  (prolog-process-insert-string (get-process "prolog")
				prolog-debug-off-string)
  (process-send-string "prolog" prolog-debug-off-string))

(defun prolog-trace-on (&optional prefix)
  "Enable tracing.
When called with prefix argument, disable tracing instead."
  (interactive "P")
  (if prefix
      (prolog-trace-off)
    (prolog-process-insert-string (get-process "prolog")
				  prolog-trace-on-string)
    (process-send-string "prolog" prolog-trace-on-string)))

(defun prolog-trace-off ()
  "Disable tracing."
  (interactive)
  (prolog-process-insert-string (get-process "prolog")
				prolog-trace-off-string)
  (process-send-string "prolog" prolog-trace-off-string))

(defun prolog-create-predicate-index ()
  "Create an index for all predicates in the buffer."
  (let ((predlist '())
	clauseinfo 
	object
	pos
	)
    (goto-char (point-min))
					; Replace with prolog-clause-start!
    (while (re-search-forward "^.+:-" nil t)
      (setq pos (match-beginning 0))
      (setq clauseinfo (prolog-clause-info))
      (setq object (prolog-in-object))
      (setq predlist (append
		      predlist
		      (list (cons 
			     (if (and (eq prolog-system 'sicstus)
				      (prolog-in-object))
				 (format "%s::%s/%d" 
					 object
					 (nth 0 clauseinfo) 
					 (nth 1 clauseinfo))
			       (format "%s/%d"
				       (nth 0 clauseinfo) 
				       (nth 1 clauseinfo)))
			     pos
			     ))))
      (prolog-goto-pred-end))
    predlist))

(defun prolog-get-predspec ()
  (save-excursion
    (let ((state (prolog-clause-info))
	  (object (prolog-in-object)))
      (if (and (eq prolog-system 'sicstus)
	       object)
	  (format "%s::%s/%d" 
		  object
		  (nth 0 state) 
		  (nth 1 state))
	(format "%s/%d"
		(nth 0 state) 
		(nth 1 state)))
      
      )))

(defun prolog-get-atom ()
  "Get atom at the current point."
  (save-excursion
    (and (re-search-forward (format "\\='?\\(%s+\\)" prolog-atom-char-regexp)
			    nil t)
	 (match-string 1))))


;; For Windows. Change backslash to slash. SICStus handles either
;; path separator but backslash must be doubled, therefore use slash.
(defun prolog-bsts (string)
  "Change backslashes to slashes in STRING."
  (let ((str1 (copy-sequence string))
	(len (length string))
	(i 0))
    (while (< i len)
      (if (char-equal (aref str1 i) ?\\)
	  (aset str1 i ?/))
      (setq i (1+ i)))
    str1))

;; For backward compatibility. Stolen from custom.el.
(or (fboundp 'match-string)
    ;; Introduced in Emacs 19.29.
    (defun match-string (num &optional string)
  "Return string of text matched by last search.
NUM specifies which parenthesized expression in the last regexp.
 Value is nil if NUMth pair didn't match, or there were less than NUM pairs.
Zero means the entire text matched by the whole regexp or whole string.
STRING should be given if the last search was by `string-match' on STRING."
  (if (match-beginning num)
      (if string
	  (substring string (match-beginning num) (match-end num))
	(buffer-substring (match-beginning num) (match-end num))))))

(defun prolog-pred-start ()
  "Return the starting point of the first clause of the current predicate."
  (save-excursion
    (goto-char (prolog-clause-start))
    ;; Find first clause, unless it was a directive
    (if (and
	 (not (looking-at ":-"))
	 (not (looking-at "[ \t]*[%/]")) ; Comment
	 )
	(let* ((pinfo (prolog-clause-info))
	       (predname (nth 0 pinfo))
	       (arity (nth 1 pinfo))
	       (op (point)))
	  (while (and (re-search-backward
		       (format "^%s\\([(\\.]\\| *%s\\)" 
			       predname prolog-head-delimiter) nil t)
		      (= arity (nth 1 (prolog-clause-info))))
	    (setq op (point)))
	  op))))

(defun prolog-pred-end ()
  "Return the position at the end of the last clause of the current predicate."
  (save-excursion
    (goto-char (prolog-clause-end))
    (goto-char (prolog-clause-start))
    (let* ((pinfo (prolog-clause-info))
	  (predname (nth 0 pinfo))
	  (arity (nth 1 pinfo))
	  (op (point)))
      (if (looking-at ":-")
	  ;; This was a directive
	  (goto-char (prolog-clause-end))
	;; It was not a directive, find the last clause
	(while (and (re-search-forward
		     (format "^%s\\([(\\.]\\| *%s\\)" 
			     predname prolog-head-delimiter) nil t)
		    (= arity (nth 1 (prolog-clause-info))))
	  (setq op (prolog-clause-end))
	  (goto-char op)))
      op)))

(defun prolog-clause-start ( &optional notallowmethods )
  "Return the position at the start of the head of the current clause.
If NOTALLOWMETHODS is non-nil then do not match on methods in
objects (relevent only if 'prolog-system' is set to 'sicstus)."
  (save-excursion
    (let ((notdone t) (retval (point-min)))
      (end-of-line)
      
      ;; SICStus object?
      (if (and (not notallowmethods)
	       (eq prolog-system 'sicstus)
	       (prolog-in-object))
	  (while (and 
		  notdone 
		  ;; Search for a head or a fact
		  (re-search-backward
		   ;; If in object, then find method start.
;		   "^[ \t]+[a-z$].*\\(:-\\|&\\|:: {\\|,\\)" 
		   "^[ \t]+[a-z$].*\\(:-\\|&\\|:: {\\)" ; The comma causes
					; problems since we cannot assume
					; that the line starts at column 0,
					; thus we don't know if the line
					; is a head or a subgoal
		   (point-min) t))
	    (if (>= (prolog-paren-balance) 0) ; To no match on "   a) :-"
		;; Start of method found
		(progn
		  (setq retval (point))
		  (setq notdone nil)))
	    )				; End of while

	;; Not in object
	(while (and 
		notdone 
		;; Search for a head or a fact
		(re-search-backward "^[a-z$']" (point-min) t))
	  ;; Start of clause found
	  (progn
	    (setq retval (point))
	    (setq notdone nil)))
      
	)

	retval)))

(defun prolog-clause-end ( &optional notallowmethods )
  "Return the position at the end of the current clause.
If NOTALLOWMETHODS is non-nil then do not match on methods in
objects (relevent only if 'prolog-system' is set to sicstus)."
  (save-excursion
    (beginning-of-line)			; Necessary since we use "^...." 
					; for the search
;(insert "q")
    (if (re-search-forward 
	 (if (and (not notallowmethods)
		  (eq prolog-system 'sicstus)
		  (prolog-in-object))
	     (format
	      "^\\(%s\\|%s\\|[^\n\'\"%%]\\)*&[ \t]*\\(\\|%%.*\\)$\\|[ \t]*}"
	      prolog-quoted-atom-regexp prolog-string-regexp)
	   (format
	    "^\\(%s\\|%s\\|[^\n\'\"%%]\\)*\\.[ \t]*\\(\\|%%.*\\)$"
	    prolog-quoted-atom-regexp prolog-string-regexp))
	 (point-max) t)
	(point)
      (error "End of clause not found!"))))

(defun prolog-clause-info ()
  "Return a (name arity) list for the current clause."
  (let (predname (arity 0))
    (save-excursion
      (goto-char (prolog-clause-start))
      (let ((op (point)))
	(skip-chars-forward "^ (\\.")
	(setq predname (buffer-substring op (point)))
	;; Retrieve the arity
	(if (looking-at prolog-left-paren)
	    (let ((endp (save-excursion
			  (prolog-forward-list) (point))))
	      (setq arity 1)
	      (forward-char 1)		; Skip the opening paren
	      (while (progn
		       (skip-chars-forward "^[({,'\"")
		       (< (point) endp))
		(if (looking-at ",")
		    (progn
		      (setq arity (1+ arity))
		      (forward-char 1)	; Skip the comma
		      )
		  ;; We found a string, list or something else we want
		  ;; to skip over. Always use prolog-tokenize,
		  ;; parse-partial-sexp does not have a 'skipover mode.
		  (goto-char (nth 5 (prolog-tokenize (point) endp 'skipover))))
		)))
	(list predname arity)
	))))

(defun prolog-in-object ()
  "Return object name if the point is inside a SICStus object definition."
  ;; Return object name if the last line that starts with a character
  ;; that is neither white space nor a comment start
  (save-excursion
    (if (save-excursion 
	  (beginning-of-line)
	  (looking-at "\\([^\n ]+\\)[ \t]*::[ \t]*{"))
	;; We were in the head of the object
	(match-string 1)
      ;; We were not in the head
      (re-search-backward "^[a-z$'}]" nil t)
      (looking-at "\\([^\n ]+\\)[ \t]*::[ \t]*{")
      (match-string 1))))

;  (save-excursion
;    (end-of-line)
;    (re-search-backward "^[a-z$'}]" nil t)
;    (looking-at "\\([^\n ]+\\)[ \t]*::[ \t]*{")
;    (match-string 1)))

(defun prolog-forward-list ()
  "Move the point to the matching right parenthesis."
  (interactive)
  (if prolog-use-prolog-tokenizer-flag
      (let ((state (prolog-tokenize (point) (point-max) 'zerodepth)))
	(goto-char (nth 5 state)))
    (forward-list)))

;; NB: This could be done more efficiently!
(defun prolog-backward-list ()
  "Move the point to the matching left parenthesis."
  (interactive)
  (if prolog-use-prolog-tokenizer-flag
      (let ((bal 0)
	    (paren-regexp (concat prolog-left-paren "\\|" prolog-right-paren))
	    (notdone t))
	(while (and notdone (re-search-backward paren-regexp nil t))
	  (cond
	   ((looking-at prolog-left-paren)
	    (if (not (prolog-in-string-or-comment))
		(setq bal (1+ bal)))
	    (if (= bal 0)
		(setq notdone nil)))
	   ((looking-at prolog-right-paren)
	    (if (not (prolog-in-string-or-comment))
		(setq bal (1- bal))))
	   )))
    (backward-list)))

(defun prolog-goto-clause-start ()
  (interactive)
  (goto-char (prolog-clause-start))
  (skip-chars-forward " \t")
)

(defun prolog-goto-clause-end ()
  (interactive)
  (goto-char (prolog-clause-end)))

(defun prolog-goto-pred-start ()
  "Go to the nearest beginning of predicate before current point.
Return the final point or nil if no such a beginning was found."
  (interactive)
  (let ((op (point))
	(pos (prolog-pred-start)))
    (if pos
	(if (= op pos)
	    (if (not (bobp))
		(progn
		  (goto-char pos)
		  (backward-char 1)
		  (setq pos (prolog-pred-start))
		  (if pos
		      (progn
			(goto-char pos)
			(point)))))
	  (goto-char pos)
	  (point)))))

;(defun prolog-goto-pred-start ()
;  (interactive)
;  (let ((op (point)))
;    (goto-char (prolog-pred-start))
;    (if (= op (point))
;	(progn
;	  (backward-char 1)
;	  (prolog-goto-pred-start)))))

(defun prolog-goto-pred-end ()
  "Go to the end of the current predicate."
  (interactive)
  (let ((op (point)))
    (goto-char (prolog-pred-end))
    (if (= op (point))
	(progn
	  (forward-line 1)
	  (prolog-goto-pred-end)))))

(defun prolog-insert-predspec ()
  "Insert the predspec for the current predicate."
  (interactive)
  (let* ((pinfo (prolog-clause-info))
	 (predname (nth 0 pinfo))
	 (arity (nth 1 pinfo)))
    (insert (format "%s/%d" predname arity))))

(defun prolog-insert-predicate-template ()
  "Insert the template for the current clause."
  (interactive)
  (let* ((n 1)
	 oldp
	 (pinfo (prolog-clause-info))
	 (predname (nth 0 pinfo))
	 (arity (nth 1 pinfo)))
    (insert predname)
    (if (> arity 0)
	(progn
	  (insert "(")
	  (setq oldp (point))
	  (while (< n arity)
	    (insert ",")
	    (setq n (1+ n)))
	  (insert ")")
	  (goto-char oldp)
	  ))
  ))

(defun prolog-insert-next-clause ()
  "Insert newline and the name of the current clause."
  (interactive)
  (insert "\n")
  (prolog-insert-predicate-template))

(defun prolog-insert-module-modeline ()
  "Insert a modeline for module specification.
This line should be first in the buffer.
The module name should be written manually just before the semi-colon."
  (interactive)
  (insert "%%% -*- Module: ; -*-\n")
  (backward-char 6))

(defun prolog-uncomment-region ( start end )
  "Uncomment the region."
  (interactive "r")
  (comment-region start end -1))

(defun prolog-goto-comment-column (&optional nocreate)
  "Move comments on the current line to the correct position.
If NOCREATE is nil (or omitted) and there is no comment on the line, then
a new comment is created."
  (interactive)
  (beginning-of-line)
  (if (or (not nocreate)
	  (and
	   (re-search-forward 
	    (format "^\\(\\(%s\\|%s\\|[^\n\'\"%%]\\)*\\)%% *" 
		    prolog-quoted-atom-regexp prolog-string-regexp)
	    (save-excursion (end-of-line) (point)) 'limit)
	   (progn
	     (goto-char (match-beginning 0))
	     (not (eq (prolog-in-string-or-comment) 'txt)))))
      (indent-for-comment)))

(defun prolog-indent-predicate ()
  "*Indent the current predicate."
  (interactive)
  (indent-region (prolog-pred-start) (prolog-pred-end) nil))

(defun prolog-indent-buffer ()
  "*Indent the entire buffer."
  (interactive)
  (indent-region (point-min) (point-max) nil))

(defun prolog-mark-clause ()
  "Put mark at the end of this clause and move point to the beginning."
  (interactive)
  (let ((pos (point)))
    (goto-char (prolog-clause-end))
    (forward-line 1)
    (beginning-of-line)
    (set-mark (point))
    (goto-char pos)
    (goto-char (prolog-clause-start))))

(defun prolog-mark-predicate ()
  "Put mark at the end of this predicate and move point to the beginning."
  (interactive)
  (let (pos)
    (goto-char (prolog-pred-end))
    (setq pos (point))
    (forward-line 1)
    (beginning-of-line)
    (set-mark (point))
    (goto-char pos)
    (goto-char (prolog-pred-start))))

;; Stolen from `cc-mode.el':
(defun prolog-electric-delete (arg)
  "Delete preceding character or whitespace.
If `prolog-hungry-delete-key-flag' is non-nil, then all preceding whitespace is
consumed.  If however an ARG is supplied, or `prolog-hungry-delete-key-flag' is
nil, or point is inside a literal then the function in the variable
`backward-delete-char' is called."
  (interactive "P")
  (if (or (not prolog-hungry-delete-key-flag)
	  arg
	  (prolog-in-string-or-comment))
      (funcall 'backward-delete-char (prefix-numeric-value arg))
    (let ((here (point)))
      (skip-chars-backward " \t\n")
      (if (/= (point) here)
	  (delete-region (point) here)
	(funcall 'backward-delete-char 1)
	))))

;; For XEmacs compatibility (suggested by Per Mildner)
(put 'prolog-electric-delete 'pending-delete 'supersede)

(defun prolog-electric-dot (arg)
  "Insert dot and newline or a head of a new clause.
If `prolog-electric-dot-flag' is nil, then simply insert dot.
Otherwise::
When invoked at the end of nonempty line, insert dot and newline.
When invoked at the end of an empty line, insert a recursive call to
the current predicate.
When invoked at the beginning of line, insert a head of a new clause
of the current predicate."
  (interactive "P")
  (if (or (not prolog-electric-dot-flag)
	  arg
	  (prolog-in-string-or-comment)
	  (save-excursion (re-search-backward "[0-9]+'\\=" nil t)))
      (funcall 'self-insert-command (prefix-numeric-value arg))
    (cond
     ;; Beginning of line
     ((bolp)
      (prolog-insert-predicate-template))
     ;; At an empty line with at least one whitespace
     ((save-excursion
	(beginning-of-line)
	(looking-at "[ \t]+$"))
      (prolog-insert-predicate-template)
      (save-excursion 
	(end-of-line)
	(insert ".\n")))
     ;; Default
     (t
      (insert ".\n"))
     )))

;;-------------------------------------------------------------------
;; Menu stuff (both for the editing buffer and for the inferior
;; prolog buffer)
;;-------------------------------------------------------------------

;; Menu for the prolog editing buffers
(defvar prolog-menu
  '(append (list (if (eq prolog-system 'mercury) "Mercury" "Prolog"))
	   (if (not (eq prolog-system 'mercury))
	       '(("Consult"
		  ["File" prolog-consult-file t]
		  ["Buffer" prolog-consult-buffer t]
		  ["Region" prolog-consult-region t]
		  ["Predicate" prolog-consult-predicate t])
		 ("Compile"
		 ["File" prolog-compile-file t]
		 ["Buffer" prolog-compile-buffer t]
		 ["Region" prolog-compile-region t]
		 ["Predicate" prolog-compile-predicate t])
		 ("Debug"
		 ["on" prolog-debug-on t]
		 ["off" prolog-debug-off t])
		 ("Trace"
		 ["on" prolog-trace-on t]
		 ["off" prolog-trace-off t])
		 ["Run Prolog" run-prolog t]
		 "-"))
	   '(["Comment region" comment-region t]
	     ["Uncomment region" prolog-uncomment-region t]
	     ["Goto comment column" indent-for-comment t]
	     "-"
	     ("Insert"
	      ["Predicate name" prolog-insert-predicate-template t]
	      ["Next clause head" prolog-insert-next-clause t]
	      ["Predicate spec." prolog-insert-predspec t]
	      ["Module modeline" prolog-insert-module-modeline t])
	     ("Goto"
	      ["Start of clause" prolog-goto-clause-start t]
	      ["End of clause" prolog-goto-clause-end t]
	      ["Start of predicate" prolog-goto-pred-start t]
	      ["End of predicate" prolog-goto-pred-end t])
	     ("Fontify"
	      ["Buffer" font-lock-fontify-buffer t])
	     ("Indent"
	      ["Line" prolog-indent-line t]
	      ["Region" indent-region t]
	      ["Predicate" prolog-indent-predicate t]
	      ["Buffer" prolog-indent-buffer t])
	     ("Mark"
	      ["Clause" prolog-mark-clause t]
	      ["Predicate" prolog-mark-predicate t]
	      ["Paragraph" mark-paragraph t])
	     )
	   (cond
	    ((eq prolog-system 'sicstus)
	     '(["Help on Predicate" prolog-help-on-predicate t]
	     ["Help on Predicate with completion"
	      prolog-find-documentation t]))
	    ((eq prolog-system 'swi)
	     '(("Help"
		["On Predicate" prolog-help-on-predicate t]
		["Apropos" prolog-help-apropos t]))))
	   )
  "The definition for the menu in the editing buffers."
  )
  
;; Menu for the inferior prolog buffer
(defvar prolog-inferior-menu
  '("Prolog"
    ("Debug"
     ["on" prolog-debug-on t]
     ["off" prolog-debug-off t])
    ("Trace"
     ["on" prolog-trace-on t]
     ["off" prolog-trace-off t])
    "-"
    ["Interrupt Prolog" comint-interrupt-subjob t]
    ["Stop Prolog" comint-stop-subjob t]
    ["Continue Prolog" comint-continue-subjob t]
    ["Quit Prolog" comint-quit-subjob t]
    ["Kill Prolog" comint-kill-subjob t]
    ["Send EOF" comint-send-eof t]
    "-"
    ["List Input History" comint-dynamic-list-input-ring t]
    ["Previous Input" comint-previous-input t]
    ["Next Input" comint-next-input t]
    ["Previous Matching Current Input"
     comint-previous-matching-input-from-input t]
    ["Next Matching Current Input"
     comint-next-matching-input-from-input t]
    ["Previous Matching Input..."
     comint-previous-matching-input t]
    ["Next Matching Input..." comint-next-matching-input t]
    ["Backward Matching Input..."
     comint-backwards-matching-input t]
    ["Forward Matching Input..."
     comint-forwards-matching-input t]
    ["Copy Old Input" comint-copy-old-input t]
    ["Kill Current Input" comint-kill-input t]
    ["Show Current Output Group" comint-show-output t]
    ["Show Maximum Output" comint-show-maximum-output t]
    ["Backward Output Group" comint-previous-prompt t]
    ["Forward Output Group" comint-next-prompt t]
    ["Kill Current Output Group" comint-kill-output t]
    ["Kill Word" backward-kill-word t]
    ["Beginning of Line" comint-bol t]
    )
  "The definition for the menu in the inferior prolog buffer."
  )

(defun prolog-menu ()
  "Add the menu in the editing buffer."
  (cond
   ((eq prolog-emacs 'gnuemacs)
    (easy-menu-define prolog-menu-map (current-local-map) ""
		      (eval prolog-menu)))
   ((eq prolog-emacs 'xemacs)
    (easy-menu-add (eval prolog-menu)))
   ;; The default
   (t
    (easy-menu-define prolog-menu-map (current-local-map) ""
		      (eval prolog-menu)))
   ))

(defun prolog-inferior-menu ()
  "Add the menu in the inferior prolog buffer."
  (cond
   ((eq prolog-emacs 'gnuemacs)
    (easy-menu-define run-prolog-menu-map (current-local-map) "" 
		      prolog-inferior-menu))
   ((eq prolog-emacs 'xemacs)
    (easy-menu-add prolog-inferior-menu))
   ;; The default
   (t
    (easy-menu-define run-prolog-menu-map (current-local-map) "" 
		      prolog-inferior-menu))
   ))

(add-hook 'prolog-mode-hook 'prolog-menu)
(add-hook 'prolog-inferior-mode-hook 'prolog-inferior-menu)

(provide 'prolog)  

;;; prolog.el ends here

